import torch
from torch import nn
import torch.nn.functional as F
from torch.utils.data import Dataset


def identity_grid(shape, stackdim, dtype=torch.float32, device="cpu"):
    """Create an nd identity grid."""
    tensors = (torch.arange(s, dtype=dtype, device=device) for s in shape)
    return torch.stack(
        torch.meshgrid(*tensors)[::-1], stackdim
    )  # z,y,x shape and flip for x, y, z coords


class Resampler(nn.Module):
    """
    Generic resampler for 2D and 3D images.
    Expects voxel coordinates as coord_field
    Args:
        input (Tensor): input batch (N x C x IH x IW) or (N x C x ID x IH x IW)
        grid (Tensor): flow-field of size (N x OH x OW x 2) or (N x OD x OH x OW x 3)
        padding_mode (str): padding mode for outside grid values
            'zeros' | 'border'. Default: 'zeros'
    """

    def __init__(
        self, coord_dim: int = 1, mode: str = "bilinear", padding_mode: str = "border"
    ):
        super().__init__()
        self.mode = mode
        self.padding_mode = padding_mode
        self.coord_dim = coord_dim

    def forward(self, input, coord_grid):
        im_shape = input.shape[2:]
        assert coord_grid.shape[self.coord_dim] == len(
            im_shape
        )  # number of coordinates should match image dimension

        coord_grid = coord_grid.movedim(self.coord_dim, -1)

        # scale for pytorch grid_sample function
        max_extent = (
            torch.tensor(
                im_shape[::-1], dtype=coord_grid.dtype, device=coord_grid.device
            )
            - 1
        )
        coord_grid = 2 * (coord_grid / max_extent) - 1
        return F.grid_sample(
            input,
            coord_grid,
            mode=self.mode,
            padding_mode=self.padding_mode,
            align_corners=True,
        )


class StableStd(torch.autograd.Function):
    @staticmethod
    def forward(ctx, tensor):
        assert tensor.numel() > 1
        ctx.tensor = tensor.detach()
        res = torch.std(tensor).detach()
        ctx.result = res.detach()
        return res

    @staticmethod
    def backward(ctx, grad_output):
        tensor = ctx.tensor.detach()
        result = ctx.result.detach()
        e = 1e-6
        assert tensor.numel() > 1
        return (
            (2.0 / (tensor.numel() - 1.0))
            * (grad_output.detach() / (result.detach() * 2 + e))
            * (tensor.detach() - tensor.mean().detach())
        )


stablestd = StableStd.apply


class ScaledTanH(nn.Module):
    def __init__(self, scaling):
        super().__init__()
        self.scaling = scaling

    def forward(self, input):
        return torch.tanh(input) * self.scaling

    def __repr__(self):
        return self.__class__.__name__ + "(" + "scaling = " + str(self.scaling) + ")"


class BiasedTanh(nn.Module):
    def __init__(self, scale_in=1.0, scale_out=1.0, bias=0.0):
        super().__init__()
        self.scale_in = scale_in
        self.scale_out = scale_out
        self.bias = bias

    def forward(self, input):
        return torch.tanh(input * self.scale_in) * self.scale_out + self.bias

    def __repr__(self):
        return (
            self.__class__.__name__
            + "("
            + "scale_in="
            + str(self.scale_in)
            + ", "
            + "scale_out="
            + str(self.scale_out)
            + ", "
            + "bias="
            + str(self.bias)
            + ")"
        )


class ScalingAF(nn.Module):
    def __init__(self, scale_factor):
        super().__init__()
        self.scale_factor = scale_factor

    def forward(self, input):
        return self.scale_factor ** torch.tanh(input)

    def __repr__(self):
        return (
            self.__class__.__name__
            + "("
            + "scale_factor="
            + str(self.scale_factor)
            + ")"
        )


class IRDataSet(Dataset):
    """
    Wrapper to convert a dataset into a dataset suitable for image registration experiments.
    """

    def __init__(self, ds, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ds = ds

    def __len__(self):
        return len(self.ds) * len(self.ds)

    def __getitem__(self, idx):
        fixed_idx = idx // len(self.ds)
        moving_idx = idx % len(self.ds)
        return {"fixed": self.ds[fixed_idx], "moving": self.ds[moving_idx]}


# Bspline functions
def bspline_kernel_nd(t, order, dtype=float, **kwargs):
    tpowers = t ** torch.arange(order, 0 - 1, -1, dtype=dtype)
    if order == 1:
        return tpowers @ torch.tensor(((-1, 1), (1, 0)), dtype=dtype)
    elif order == 2:
        return (
            tpowers
            @ torch.tensor(((1, -2, 1), (-2, 2, 0), (1, 1, 0)), dtype=dtype)
            / 2.0
        )
    elif order == 3:
        return (
            tpowers
            @ torch.tensor(
                ((-1, 3, -3, 1), (3, -6, 3, 0), (-3, 0, 3, 0), (1, 4, 1, 0)),
                dtype=dtype,
            )
            / 6.0
        )


def bspline_convolution_kernel(upsampling_factors, order, dtype=float):
    ndim = len(upsampling_factors)
    for i, us_factor in enumerate(upsampling_factors):
        t = torch.linspace(1 - (1 / us_factor), 0, us_factor)
        ker1D = bspline_kernel_nd(t[:, None], order, dtype).T.flatten()
        shape = (1,) * i + ker1D.shape + (1,) * (ndim - 1 - i)
        try:
            kernel = kernel * ker1D.view(shape)
        except NameError:
            kernel = ker1D.view(shape)
    return kernel


# transform
def scaling_matrix_eff(x, ndim=2):
    assert (
        x.ndim == 2
    ), "Input should be a tensor of (m, n), where m is number of instances, and n number of parameters."
    assert ndim in (2, 3), "Only 2D and 3D implemented."
    assert ndim == x.shape[-1], f"Number of parameters for {ndim}D should be {ndim}."
    dtype, device = x.dtype, x.device
    T = torch.zeros((len(x), ndim, ndim), dtype=dtype, device=device)
    sel_mask = torch.eye(ndim, device=device, dtype=torch.bool)
    T[:, sel_mask] = x
    return T


def rotation_matrix_eff(x, axis=0, ndim=2):
    """
    For 3D axis = x: 2, y: 1, z: 0.
    """
    assert (
        x.ndim == 2 and x.shape[-1] == 1
    ), "Input should be a tensor of (m, 1), where m is number of instances."
    assert ndim in (2, 3), "Only 2D and 3D implemented."
    dtype, device = x.dtype, x.device
    T = torch.eye(ndim, dtype=dtype, device=device)[None].repeat(len(x), 1, 1)
    lidx, hidx = ((0, 1), (0, 2), (1, 2))[axis]
    c = torch.cos(x)
    s = torch.sin(x)
    T[:, lidx, lidx] = c.squeeze()
    T[:, lidx, hidx] = s.squeeze()
    T[:, hidx, lidx] = -s.squeeze()
    T[:, hidx, hidx] = c.squeeze()
    return T


def shearing_matrix_eff(x, ndim=2):
    assert (
        x.ndim == 2
    ), "Input should be a tensor of (m, n), where m is number of instances, and n number of parameters."
    assert ndim in (2, 3), "Only 2D and 3D implemented."
    assert (ndim - 1) * ndim == x.shape[
        -1
    ], f"Number of parameters for {ndim}D should be {(ndim - 1) * ndim}"
    dtype, device = x.dtype, x.device
    T = torch.eye(ndim, dtype=dtype, device=device)[None].repeat(len(x), 1, 1)
    T[:, ~torch.eye(ndim, device=device, dtype=torch.bool)] = torch.tan(x)
    return T


def translation_matrix(x, ndim=2):  # not used for efficient transforms
    assert (
        x.ndim == 2
    ), "Input should be a tensor of (m, n), where m is number of instances, and n number of parameters."
    assert ndim in (2, 3), "Only 2D and 3D implemented."
    assert ndim == x.shape[-1], f"Number of parameters for {ndim}D should be {ndim}."
    dtype, device = x.dtype, x.device
    T = torch.eye(ndim + 1, dtype=dtype, device=device)[None].repeat(len(x), 1, 1)
    T[:, :ndim, ndim] = x
    return T


def tmat_2d_affine(rotation, scale, shear):
    c = torch.cos(rotation[:, 0])
    s = torch.sin(rotation[:, 0])
    sx = scale[:, 0]
    sy = scale[:, 1]
    sh = torch.tan(shear[:, 0])
    T = torch.stack((sx * (c - s * sh), -s * sy, (sx * (s + c * sh)), sy * c)).view(
        -1, 2, 2
    )
    return T


shearmat2d = shearing_matrix_eff
rotmat2d = rotation_matrix_eff
scalemat2d = scaling_matrix_eff


def batch_transform_efficient(
    coord_grid, coord_dim, Tmat, translation, Forigin, Morigin
):
    # print(coord_grid.shape)
    coord_grid = coord_grid.movedim(coord_dim, -1)
    # print(coord_grid.shape)
    shape = coord_grid.shape
    coord_grid = coord_grid.view(shape[0], -1, shape[-1])  # flatten
    # print(coord_grid.shape, Forigin.shape)
    coord_grid = coord_grid + Forigin[:, None]
    coord_grid = coord_grid @ Tmat.transpose(
        2, 1
    )  # transpose because of switched order
    coord_grid = coord_grid + (translation[:, None] - Morigin[:, None])
    shape = (len(coord_grid),) + shape[1:]  # enable broadcasting
    coord_grid = coord_grid.view(shape)  # recover original shape
    coord_grid = coord_grid.movedim(-1, coord_dim)
    return coord_grid


def batch_transform(coord_grid, coord_dim, Tmat, Forigin, Morigin):
    coord_grid = coord_grid.movedim(coord_dim, -1)
    ndim = coord_grid.shape[-1]
    Tmat = translation_matrix(-Morigin) @ Tmat @ translation_matrix(Forigin)
    coord_grid = torch.cat(
        (
            coord_grid,
            torch.ones(
                coord_grid.shape[:-1] + (1,),
                dtype=coord_grid.dtype,
                device=coord_grid.device,
            ),
        ),
        dim=ndim + 1,
    )  # add extra dimension for MM
    shape = coord_grid.shape
    coord_grid = coord_grid.view(shape[0], -1, shape[-1])  # flatten
    coord_grid = coord_grid @ Tmat.transpose(2, 1)  # MM
    coord_grid = coord_grid.view(shape)  # recover shape
    coord_grid = coord_grid.narrow(ndim + 1, 0, ndim)  # remove extra dimension
    return coord_grid.movedim(-1, coord_dim)



from torch import nn, Tensor

from torchir.utils import Resampler
from torchir.transformers import Transformer


class DLIRFramework(nn.Module):
    """Deep Learning Image Registration (DLIR) Framework, a pipeline for transformers.

    The framework is an adaptation of the code presented in `De Vos 2019 <https://www.sciencedirect.com/science/article/pii/S1361841518300495>`_.

    Note: The framework differs from the original paper in the sense that DVFs are not combined by
    addition, but by composition which should result in an increase of registration precision in
    most cases `Hering et al. 2019 <https://link.springer.com/chapter/10.1007/978-3-030-32226-7_29>`_.
    """

    def __init__(self, only_last_trainable: bool = True, return_field: bool = False):
        super().__init__()
        self._resampler = Resampler(coord_dim=-1)
        self.stages = nn.ModuleList()
        self.only_last_trainable = only_last_trainable
        self.return_field = return_field

    def add_stage(self, network: nn.Module, transformer: Transformer):
        if self.only_last_trainable and len(self.stages) > 0:
            self.stages[-1].eval()
            for param in self.stages[-1].parameters():
                param.requires_grad = False
        stage = nn.ModuleDict({"network": network, "transformer": transformer})
        self.stages.append(stage)

    def forward(self, fixed: Tensor, moving: Tensor) -> Tensor:
        coord_grid = None
        warped = moving
        for stage in self.stages:
            parameters = stage["network"](fixed, warped)
            warped, coord_grid = stage["transformer"](
                parameters, fixed, moving, coord_grid, return_coordinate_grid=True
            )
        return warped



import torch
from torch import Tensor

from torchir.utils import identity_grid


def bending_energy_3d(
    coord_grid: Tensor, vector_dim: int = -1, dvf_input: bool = False
) -> Tensor:
    """Calculates bending energy penalty for a 3D coordinate grid.

    For further details regarding this regularization please read the work by `Rueckert 1999`_.

    Args:
        coord_grid: 3D coordinate grid, i.e. a 5D Tensor with standard dimensions
        (n_samples, 3, z, y, x).
        vector_dim: Specifies the location of the vector dimension. Default: -1
        dvf_input: If ``True``, coord_grid is assumed a displacement vector field and
        an identity_grid will be added. Default: ``False``

    Returns:
        Bending energy per instance in the batch.

    .. _Rueckert 1999: https://ieeexplore.ieee.org/document/796284

    """
    assert coord_grid.ndim == 5, "Input tensor should be 5D, i.e. 3D images."

    if vector_dim != 1:
        coord_grid = coord_grid.movedim(vector_dim, -1)

    if dvf_input:
        coord_grid = coord_grid + identity_grid(coord_grid.shape[2:], stackdim=0)

    d_z = torch.diff(coord_grid, dim=1)
    d_y = torch.diff(coord_grid, dim=2)
    d_x = torch.diff(coord_grid, dim=3)

    d_zz = torch.diff(d_z, dim=1)[:, :, :-2, :-2]
    d_zy = torch.diff(d_z, dim=2)[:, :-1, :-1, :-2]
    d_zx = torch.diff(d_z, dim=3)[:, :-1, :-2, :-1]
    d_yy = torch.diff(d_y, dim=2)[:, :-2, :, :-2]
    d_yx = torch.diff(d_y, dim=3)[:, :-2, :-1, :-1]
    d_xx = torch.diff(d_x, dim=3)[:, :-2, :-2, :]

    return torch.mean(
        d_zz ** 2 + d_yy ** 2 + d_xx ** 2 + 2 * (d_zy ** 2 + d_zx ** 2 + d_yx ** 2),
        axis=(1, 2, 3, 4),
    )


def bending_energy_2d(
    coord_grid: Tensor, vector_dim: int = -1, dvf_input: bool = False
) -> Tensor:
    """Calculates bending energy penalty for a 2D coordinate grid.

    For further details regarding this regularization please read the work by `Rueckert 1999`_.

    Args:
        coord_grid: 2D coordinate grid, i.e. a 4D Tensor with standard dimensions
        (n_samples, 2, y, x).
        vector_dim: Specifies the location of the vector dimension. Default: -1
        dvf_input: If ``True``, coord_grid is assumed a displacement vector field and
        an identity_grid will be added. Default: ``False``

    Returns:
        Bending energy per instance in the batch.

    .. _Rueckert 1999: https://ieeexplore.ieee.org/document/796284

    """
    assert coord_grid.ndim == 4, "Input tensor should be 4D, i.e. 2D images."

    if vector_dim != 1:
        coord_grid = coord_grid.movedim(vector_dim, -1)

    if dvf_input:
        coord_grid = coord_grid + identity_grid(coord_grid.shape[2:], stackdim=0)

    d_y = torch.diff(coord_grid, dim=1)
    d_x = torch.diff(coord_grid, dim=2)

    d_yy = torch.diff(d_y, dim=1)[:, :, :-2]
    d_yx = torch.diff(d_y, dim=2)[:, :-1, :-1]
    d_xx = torch.diff(d_x, dim=2)[:, :-2, :]

    return torch.mean(d_yy ** 2 + d_xx ** 2 + 2 * d_yx ** 2, axis=(1, 2, 3))



import math
from typing import Optional, Tuple

import torch
from torch import Tensor
from torch.nn.modules.loss import _Loss

from torchir.utils import stablestd


# TODO: add masks.
def ncc(x1, x2, e=1e-10):
    assert x1.shape == x2.shape, "Inputs are not of equal shape"
    cc = ((x1 - x1.mean()) * (x2 - x2.mean())).mean()
    std = stablestd(x1) * stablestd(x2)
    ncc = cc / (std + e)
    return ncc


def ncc_mask(x1, x2, mask, e=1e-10):  # TODO: calculate ncc per sample
    assert x1.shape == x2.shape, "Inputs are not of equal shape"
    x1 = torch.masked_select(x1, mask)
    x2 = torch.masked_select(x2, mask)
    cc = ((x1 - x1.mean()) * (x2 - x2.mean())).mean()
    std = stablestd(x1) * stablestd(x2)
    ncc = cc / (std + e)
    return ncc


class NCC(_Loss):
    def __init__(self, use_mask: bool = False):
        super().__init__()
        if use_mask:
            self.forward = self.masked_metric
        else:
            self.forward = self.metric

    def metric(self, fixed: Tensor, warped: Tensor) -> Tensor:
        return -ncc(fixed, warped)

    def masked_metric(self, fixed: Tensor, warped: Tensor, mask: Tensor) -> Tensor:
        return -ncc_mask(fixed, warped, mask)


def nmi_gauss(x1, x2, x1_bins, x2_bins, sigma=1e-3, e=1e-10):
    assert x1.shape == x2.shape, "Inputs are not of similar shape"

    def gaussian_window(x, bins, sigma):
        assert x.ndim == 2, "Input tensor should be 2-dimensional."
        return torch.exp(
            -((x[:, None, :] - bins[None, :, None]) ** 2) / (2 * sigma ** 2)
        ) / (math.sqrt(2 * math.pi) * sigma)

    x1_windowed = gaussian_window(x1.flatten(1), x1_bins, sigma)
    x2_windowed = gaussian_window(x2.flatten(1), x2_bins, sigma)
    p_XY = torch.bmm(x1_windowed, x2_windowed.transpose(1, 2))
    p_XY = p_XY + e  # deal with numerical instability

    p_XY = p_XY / p_XY.sum((1, 2))[:, None, None]

    p_X = p_XY.sum(1)
    p_Y = p_XY.sum(2)

    I = (p_XY * torch.log(p_XY / (p_X[:, None] * p_Y[:, :, None]))).sum((1, 2))

    marg_ent_0 = (p_X * torch.log(p_X)).sum(1)
    marg_ent_1 = (p_Y * torch.log(p_Y)).sum(1)

    normalized = -1 * 2 * I / (marg_ent_0 + marg_ent_1)  # harmonic mean

    return normalized


def nmi_gauss_mask(x1, x2, x1_bins, x2_bins, mask, sigma=1e-3, e=1e-10):
    def gaussian_window_mask(x, bins, sigma):

        assert x.ndim == 1, "Input tensor should be 2-dimensional."
        return torch.exp(-((x[None, :] - bins[:, None]) ** 2) / (2 * sigma ** 2)) / (
            math.sqrt(2 * math.pi) * sigma
        )

    x1_windowed = gaussian_window_mask(torch.masked_select(x1, mask), x1_bins, sigma)
    x2_windowed = gaussian_window_mask(torch.masked_select(x2, mask), x2_bins, sigma)
    p_XY = torch.mm(x1_windowed, x2_windowed.transpose(0, 1))
    p_XY = p_XY + e  # deal with numerical instability

    p_XY = p_XY / p_XY.sum()

    p_X = p_XY.sum(0)
    p_Y = p_XY.sum(1)

    I = (p_XY * torch.log(p_XY / (p_X[None] * p_Y[:, None]))).sum()

    marg_ent_0 = (p_X * torch.log(p_X)).sum()
    marg_ent_1 = (p_Y * torch.log(p_Y)).sum()

    normalized = -1 * 2 * I / (marg_ent_0 + marg_ent_1)  # harmonic mean

    return normalized


class NMI(_Loss):
    """Normalized mutual information metric.

    As presented in the work by `De Vos 2020: <https://www.spiedigitallibrary.org/conference-proceedings-of-spie/11313/113130R/Mutual-information-for-unsupervised-deep-learning-image-registration/10.1117/12.2549729.full?SSO=1>`_

    """

    def __init__(
        self,
        intensity_range: Optional[Tuple[float, float]] = None,
        nbins: int = 32,
        sigma: float = 0.1,
        use_mask: bool = False,
    ):
        super().__init__()
        self.intensity_range = intensity_range
        self.nbins = nbins
        self.sigma = sigma
        if use_mask:
            self.forward = self.masked_metric
        else:
            self.forward = self.metric

    def metric(self, fixed: Tensor, warped: Tensor) -> Tensor:
        with torch.no_grad():
            if self.intensity_range:
                fixed_range = self.intensity_range
                warped_range = self.intensity_range
            else:
                fixed_range = fixed.min(), fixed.max()
                warped_range = warped.min(), warped.max()

        bins_fixed = torch.linspace(
            fixed_range[0],
            fixed_range[1],
            self.nbins,
            dtype=fixed.dtype,
            device=fixed.device,
        )
        bins_warped = torch.linspace(
            warped_range[0],
            warped_range[1],
            self.nbins,
            dtype=fixed.dtype,
            device=fixed.device,
        )

        return -nmi_gauss(
            fixed, warped, bins_fixed, bins_warped, sigma=self.sigma
        ).mean()

    def masked_metric(self, fixed: Tensor, warped: Tensor, mask: Tensor) -> Tensor:
        with torch.no_grad():
            if self.intensity_range:
                fixed_range = self.intensity_range
                warped_range = self.intensity_range
            else:
                fixed_range = fixed.min(), fixed.max()
                warped_range = warped.min(), warped.max()

        bins_fixed = torch.linspace(
            fixed_range[0],
            fixed_range[1],
            self.nbins,
            dtype=fixed.dtype,
            device=fixed.device,
        )
        bins_warped = torch.linspace(
            warped_range[0],
            warped_range[1],
            self.nbins,
            dtype=fixed.dtype,
            device=fixed.device,
        )

        return -nmi_gauss_mask(
            fixed, warped, bins_fixed, bins_warped, mask, sigma=self.sigma
        )



from abc import ABC, abstractmethod
from typing import Optional, Tuple, Union

import numpy as np
import torch
from torch import nn, Tensor, Size
import torch.nn.functional as F

from torchir.utils import identity_grid, Resampler
from torchir.utils import bspline_convolution_kernel
from torchir.utils import (
    batch_transform_efficient,
    rotation_matrix_eff,
    shearing_matrix_eff,
    scaling_matrix_eff,
)


class Transformer(ABC, nn.Module):
    def __init__(self, ndim: int, coord_dim: int = 1):
        super().__init__()
        self.ndim = ndim
        self.coord_dim = coord_dim
        self._resampler = Resampler(coord_dim=coord_dim)

    @abstractmethod
    def apply_transform(
        self,
        parameters: Tensor,
        fixed_image: Tensor,
        moving_image: Tensor,
        coordinate_grid: Optional[Tensor] = None,
    ) -> Tensor:
        """apply the parameters to get the transformed coord_grid"""
        pass

    def forward(
        self,
        parameters: Tensor,
        fixed_image: Tensor,
        moving_image: Tensor,
        coordinate_grid: Optional[Tensor] = None,
        return_coordinate_grid: bool = False,
    ) -> Tuple[Tensor]:

        coordinate_grid = self.apply_transform(
            parameters, fixed_image, moving_image, coordinate_grid
        )

        ret = self._resampler(moving_image, coordinate_grid)
        if return_coordinate_grid:
            ret = (ret, coordinate_grid)
        return ret


class BsplineTransformer(Transformer):
    """
    third order bspline. upsampling via transposed convolutions.
    """

    def __init__(
        self,
        *args,
        upsampling_factors: Tuple[int],
        order: int = 3,
        trainable: bool = False,
        apply_diffeomorphic_limits: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.upsampling_factors = upsampling_factors
        ConvTranspose = (nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)[
            self.ndim - 1
        ]

        bspline_kernel = self.make_bspline_kernel(self.upsampling_factors, order=order)
        kernel_size = bspline_kernel.shape
        crop_size = tuple(int(el * 3 / 8) if el != 5 else 2 for el in kernel_size)
        upsampler = ConvTranspose(
            1,
            1,
            kernel_size,
            stride=self.upsampling_factors,
            padding=crop_size,
            bias=False,
        )
        upsampler.weight = nn.Parameter(
            bspline_kernel[None, None], requires_grad=trainable
        )

        self._upsampler = upsampler
        self.apply_diffeomorphic_limits = apply_diffeomorphic_limits

    @staticmethod
    def make_bspline_kernel(upsampling_factors, order, dtype=torch.float32):
        # TODO: solve the issues with kernel shapes
        bspline_kernel = bspline_convolution_kernel(
            upsampling_factors, order=order, dtype=dtype
        )

        if (np.array(bspline_kernel.shape[::-1]) == 4).any() or (
            np.array(bspline_kernel.shape[::-1]) == 2
        ).any():  # hack to deal with 1 strides and kernel size of 4
            padding = list()
            for s in bspline_kernel.shape[::-1]:
                if s == 4 or s == 2:
                    padding.extend([1, 0])
                else:
                    padding.extend([0, 0])
            bspline_kernel = F.pad(bspline_kernel, padding, mode="constant")

        return bspline_kernel

    def create_dvf(self, bspline_parameters, output_shape):
        assert bspline_parameters.shape[1] == self.ndim
        shape = bspline_parameters.shape
        dvf = self._upsampler(
            bspline_parameters.view((shape[0] * self.ndim, 1) + shape[2:]),
            output_size=output_shape,
        )
        newshape = dvf.shape

        return dvf.view((shape[0], self.ndim) + newshape[2:])

    def apply_transform(
        self,
        bspline_coefficients: Tensor,
        fixed_image: Tensor,
        moving_image: Tensor,
        coordinate_grid: Tensor = None,
    ):
        device = bspline_coefficients.device
        dtype = bspline_coefficients.dtype
        if self.apply_diffeomorphic_limits:
            bspline_coefficients = torch.tanh(bspline_coefficients) * (
                0.4
                * torch.tensor(
                    self.upsampling_factors[::-1],
                    dtype=dtype,
                    device=device,
                ).view((1, self.ndim) + (1,) * self.ndim)
            )

        shape = fixed_image.shape[2:]
        dvf = self.create_dvf(bspline_coefficients, output_shape=shape)
        new_grid = dvf + identity_grid(
            dvf.shape[2:], stackdim=0, dtype=dtype, device=device
        )
        if coordinate_grid is not None:
            assert shape == coordinate_grid.shape[2:]
            new_grid = self._resampler(coordinate_grid, new_grid)
        return new_grid


class AffineTransformer(Transformer):
    def apply_transform(
        self,
        parameters: Tuple[Tensor],
        fixed_image: Tensor,
        moving_image: Tensor,
        coordinate_grid: Optional[Tensor] = None,
    ) -> Tensor:
        """
        :param parameters: translation, rotation, scale, shear
        :param fixed_image:
        :param moving_image:
        :param coordinate_grid:
        :return:

        Note: If ndim == 3, number of parameters are 3 for translation, rotation, and scale; and
        6 for shear. If ndim == 2, number of parameters are 2 for translation, scale, and shear;
        and 1 for rotation
        """
        translation, rotation, scale, shear = parameters

        if self.ndim == 2:
            rot_mat = rotation_matrix_eff(rotation, ndim=self.ndim)
        elif self.ndim == 3:
            rot_mat = (
                rotation_matrix_eff(rotation[:, 0], axis=2, ndim=self.ndim)
                @ rotation_matrix_eff(rotation[:, 1], axis=1, ndim=self.ndim)
                @ rotation_matrix_eff(rotation[:, 2], axis=0, ndim=self.ndim)
            )

        Tmat = (
            scaling_matrix_eff(scale, ndim=self.ndim)
            @ shearing_matrix_eff(shear, ndim=self.ndim)
            @ rot_mat
        )

        f_origin = -(
            torch.tensor(
                fixed_image.shape[2:],
                dtype=fixed_image.dtype,
                device=fixed_image.device,
            )[None]
            / 2
        )
        m_origin = -(
            torch.tensor(
                moving_image.shape[2:],
                dtype=moving_image.dtype,
                device=moving_image.device,
            )[None]
            / 2
        )

        if coordinate_grid is None:
            coordinate_grid = identity_grid(
                fixed_image.shape[2:],
                stackdim=0,
                dtype=fixed_image.dtype,
                device=fixed_image.device,
            )[None].movedim(1, self.coord_dim)

        coordinate_grid = batch_transform_efficient(
            coordinate_grid, self.coord_dim, Tmat, translation, f_origin, m_origin
        )

        return coordinate_grid


class DirectDVFTransformer(Transformer):
    def apply_transform(
        self,
        dvf: Tensor,
        fixed_image: Tensor,
        moving_image: Tensor,
        coordinate_grid: Optional[Tensor] = None,
    ) -> Tensor:
        return dvf


class DiffeomorphicFlowTransformer(Transformer):
    """
    Implementation for diffeormorphic DVFs as used in `Voxelmorph <https://github.com/voxelmorph/voxelmorph>`_.
    Uses scaling and squaring of a velocity field to obtain a diffeomorphic DVF.
    """

    def __init__(self, *args, nsteps=8, **kwargs):
        super().__init__(*args, **kwargs)

        assert nsteps >= 0, f"nsteps should be >= 0"
        self.nsteps = nsteps
        self.scale = 1.0 / (2 ** self.nsteps)

    def scaling_squaring(self, flow_field):
        flow_field = flow_field * self.scale
        for _ in range(self.nsteps):
            grid = identity_grid(
                flow_field.shape[2:],
                stackdim=0,
                device=flow_field.device,
                dtype=flow_field.dtype,
            )[None].movedim(1, self.coord_dim)
            flow_field = flow_field + self._resampler(flow_field, flow_field + grid)
        return flow_field

    def apply_transform(
        self,
        flow_field: Tensor,
        fixed_image: Tensor,
        moving_image: Tensor,
        coordinate_grid: Optional[Tensor] = None,
    ) -> Tensor:
        dvf = self.scaling_squaring(flow_field)
        device = dvf.device
        dtype = dvf.dtype
        new_grid = dvf + identity_grid(
            dvf.shape[2:], stackdim=0, dtype=dtype, device=device
        )
        if coordinate_grid:
            assert shape == coordinate_grid.shape[2:]
            new_grid = self._resampler(coordinate_grid, new_grid)
        return new_grid



from distutils.core import setup

setup(
    name="torchir",
    packages=["torchir"],  # Chose the same as "name"
    version="0.1",
    license="MIT",
    description="Pytorch Image Registration Library",
    author="Bob D. de Vos",  # Type in your name
    author_email="b.d.devos@amsterdamumc.nl",
    url="https://github.com/BDdeVos/TorchIR",
    download_url="https://github.com/BDdeVos/TorchIR/archive/refs/tags/version_01.tar.gz",  # I explain this later on
    keywords=[
        "PyTorch",
        "Deep Learning",
        "Image Registration",
        "Affine Registration",
        "Deformable Registration",
    ],  # Keywords that define your package best
    install_requires=[  # I get to this in a second
        "torch",
        "numpy",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",  # Define that your audience are developers
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
    ],
)
